/*************************************************************
 *
 * This is a part of the Aliyun RTS Service SDK.
 * Copyright (C) 2021 alibaba
 * All rights reserved.
 *
 *************************************************************/

#ifndef __ALI_RTS_SDK_H
#define __ALI_RTS_SDK_H
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" {
#endif

#if defined(_MSC_VER)
#if defined(__BUILDING_ALIRTS_SDK__)
#define __alirts_api_ __declspec(dllexport)
#else
#define __alirts_api_ __declspec(dllimport)
#endif
#else
#if defined(__BUILDING_ALIRTS_SDK__)
#define __alirts_api_ __attribute__((visibility("default")))
#else
#define __alirts_api_
#endif
#endif

/**
 * The definition of alirts handler
 */
typedef void alirts_handler_t;

/**
 * The definition of log level enum
 * 
 * @note default ALIRTS_LOG_INFO
 */
typedef enum {
	ALIRTS_LOG_ERROR = 1,
	ALIRTS_LOG_WARN,
	ALIRTS_LOG_INFO,
	ALIRTS_LOG_DEBUG,
} alirts_log_level_e;

/**
 * The definition of the alirts_err_code_e enum.
 */
typedef enum {
  /** 
   * No error 
   */
	NONE_ERR = 0,

  /** 
   * Video encoder init error
   */
  VIDEO_ENCODER_INIT_ERR = 100,

  /** 
   * Video encoder encode error 
   */
  VIDEO_ENCODER_ENCODE_ERR = 101,

  /** 
   * Audio encoder init error
   */
  AUDIO_ENCODER_INIT_ERR = 200,

  /** 
   * Audio encoder encode error
   */
  AUDIO_ENCODER_ENCODE_ERR = 201,

  /** 
   * Input config parameter error
   */
  CONFIG_PARAM_ERR = 300,

} alirts_err_code_e;


/**
 * The definition of the alirts_media_type_e enum.
 * 
 * @note Audio only support PCM
 *       Video only support I420
 */
typedef enum {
  /** 
   * Input video formate, now only I420
   */
  MEDIA_TYPE_VIDEO_I420,

  /** 
   * Input audio formate, now only pcm
   */
  MEDIA_TYPE_AUDIO_PCM,
} alirts_media_type_e;


/**
 * The definition of the alirts_audio_format_e enum.
 * 
 * @note Now only support FORMAT_16BIT
 */
typedef enum {
  /** 
   * Audio 16bit formate
   */
	FORMAT_16BIT = 16,
  /** 
   * Audio 32bit formate
   * @note not support now
   */
	FORMAT_32BIT = 32,
} alirts_audio_format_e;

/**
 * The definition of the alirts_audio_sample_rate_e enum.
 * 
 * @note Now only support 44.1k and 48K
 */
typedef enum {
	SAMPLE_RATE_44100 = 44100,
	SAMPLE_RATE_48000 = 48000,
} alirts_audio_sample_rate_e;

/**
 * The definition of the alirts_video_codec_e enum.
 * 
 * @note 1.Now only support H264 and H265
 *       2.When H265 is enabled, if the performance of the computer is poor, will switch to H264.
 */
typedef enum {
	VIDEO_CODEC_H264 = 1,
	VIDEO_CODEC_H265 = 2,
} alirts_video_codec_e;

/**
 * The definition of the alirts_video_encoder_preset_e enum.
 */
typedef enum {
  VIDEO_PRESET_VERYFAST = 1,
  VIDEO_PRESET_FAST = 2,
  VIDEO_PRESET_MEDIUM = 3,
  VIDEO_PRESET_SLOW = 4,
} alirts_video_encoder_preset_e;

/**
 * The definition of the alirts_video_encoder_profile_e enum.
 */
typedef enum {
  VIDEO_PROFILE_BASELINE = 1,
  VIDEO_PROFILE_MAIN = 2,
  VIDEO_PROFILE_HIGH = 3,
} alirts_video_encoder_profile_e;

/**
 * Alirts config
 */
typedef struct {

  /** 
   * Video width
   */
  int width;

  /** 
   * Video height
   */
  int height;

  /** 
   * Video gop size (seconds)
   * 
   * @note 0s <= gop_size <= 30s
   */
  int gop_size;

  /** 
   * Video fps
   */
  int max_fps;

  /** 
   * Max video bitrate
   */
  int video_bitrate_bps;

  /** 
   * Video codec type
   */
  alirts_video_codec_e video_codec;

  /** 
   * Video encoder preset
   */
  alirts_video_encoder_preset_e video_preset;

  /** 
   * Video encoder profile
   */
  alirts_video_encoder_profile_e video_profile;


  /** 
   * Audio sample rate
   * 
   * @note Only support 44100 and 48000
   */
  alirts_audio_sample_rate_e sample_rate;

  /** 
   * Audio sample formate
   * 
   * @note Only support FORMAT_16BIT
   */
  alirts_audio_format_e sample_format;

  /** 
   * Audio channels
   */
  int channel;

  /** 
   * Max audio bitrate
   */
  int audio_bitrate_bps;

  /** 
   * Enable auto reconnect
   * 
   * @note Enable automatic reconnection is recommended
   */
  bool enable_auto_reconnect;

  /** 
   * Reconnect timeout time (seconds)
   * 
   * @note if enable auto reconnect, set the timeout time.
   *       10s is recommended
   */
  int reconnect_timeout;

  /** 
   * push url
   */
  char* url;

} alirts_config_t;

/** 
 * Push statistics info
 */
typedef struct {
  /** 
   * Estimate send bandwaidth
   */
  int available_send_bandwidth_bps;

  /** 
   * Set video encoder target bitrate
   */
  int target_enc_bitrate_bps;

  /** 
   * Video encoder actual encode bitrate
   */
  int actual_enc_bitrate_bps;

  /** 
   * Video and audio send bitrate
   */
  int current_send_bitrate_bps;

  /** 
   * Video send bitrate
   */
  int current_video_send_bitrate_bps;

  /** 
   * Video retransmission bitrate
   */
  int current_video_resend_bitrate_bps;

  /** 
   * Audio send bitrate
   */
  int current_audio_send_bitrate_bps;

  /** 
   * Audio retransmission bitrate
   */
  int current_audio_resend_bitrate_bps;

  /** 
   * Average video encode pnsr
   */
  int current_video_enc_psnr;

  /** 
   * Total send bytes since start push
   */
  int total_send_bytes;
  
  /** 
   * Video send lost (%)
   */
  int video_current_sent_lost;

  /** 
   * Audio send lost (%)
   */
  int audio_current_sent_lost;

  /** 
   * Round trip delay (ms)
   */
  int rtt_ms;

} alirts_push_info_t;


/** 
 * Media frame 
 */
typedef struct {
  /** 
   * Media frame type
   */
  alirts_media_type_e type;

  /** 
   * Media frame buffer
   * 
   * @note audio only used data[0]
   */
  uint8_t* data[4];

  /** 
   * Media frame linesize
   * 
   * @note audio only used linesize[0]
   */
  int linesize[4];

  /** 
   * Video frame width
   */
  int width;

  /** 
   * Video frame height
   */
  int height;

  /** 
   * audio frame samples
   */
  int audio_nb_samples;

  /** 
   * audio sample rate
   */
  int audio_sample_rate;

  /** 
   * audio channels
   */
  int audio_channels;

  /** 
   * Media frame timestamp
   */
  int64_t timestamp;

} alirts_frame_t;


/**
 * alirts sdk event handler
 */
typedef struct {

  // User data, this value will be along with other callbacks like on_error
  void* user_data;

  /**
   * Callback the sdk log.
   *
   * @param[in] level  Log level, see #alirts_log_level_e
   * @param[in] log    log data
   */
  void (*on_log_data)(int level, const char *log);

  /**
   * Report error message during runtime.
   *
   * @param[in] user_data      User input user data
   * @param[in] code           Error code, see #alirts_err_code_e
   * @param[in] msg            Error message
   */
  void (*on_error)(void* user_data, int code, const char *msg);

  /**
   * Occurs when connect to server successfully,
   * can send audio and video data now
   *
   * @param[in] user_data    User input user data
   * 
   * @note alirts_send_frame must call after this event was triggered
   */
  void (*on_push_start)(void* user_data);

  /**
    * Occurs when stop pushing stream successfully
    *
    * @param[in] user_data    User input user data
    */
  void (*on_push_stop)(void* user_data);

  /**
   * Occurs when connection is disconnected from the server.
   *
   * @param[in] user_data    User input user data
   */
  void (*on_connection_lost)(void* user_data);

  /**
   * Occurs when reconnect state changed (success or failed) after disconnect
   *
   * When connection is disconnected from the server due to network problems,
   * SDK will retry to connect automatically. If success or failed with timeout, it will be triggered.
   *
   * @param[in] user_data    User input user data
   * @param[in] success      true:reconnet success; false: reconnet failed with timeout
   */
  void (*on_reconnect_state)(void* user_data, bool success);

} alirts_event_handler_t;



/////////////////////////////////////////////////////////////////////////////////
//
//                         SDK API
//
// Please note the call order:
//   1. alirts_create;
//   2. alirts_start;
//   3. alirts_send_frame;
//   4. alirts_stop;
//   5. alirts_destroy;
// The wrong call order will cause the SDK to behave abnormally
/////////////////////////////////////////////////////////////////////////////////

/**
 * Get SDK version.
 *
 * @return A const static string describes the SDK version
 */
extern __alirts_api_ const char *alirts_get_version();

/**
 * Set the log level.
 *
 * @param[in] level Log level. see#alirts_log_level_e
 *
 * @return
 * - = 0: Success
 * - < 0: Failure
 */
extern __alirts_api_ int alirts_set_log_level(alirts_log_level_e level);

/**
 * Create the alirts service
 *
 * @param[in] event_handler event handler, see#alirts_event_handler_t
 *
 * @return
 * - NULL: Failure
 * - non-NULL: alirts handler see#alirts_handler_t
 */
extern __alirts_api_ alirts_handler_t *alirts_create(const alirts_event_handler_t *event_handler);


/**
 * Start push stream
 *
 * @param[in] handler alirts handler
 * @param[in] config  alirts config, see#alirts_config_t
 *                    config must set, otherwise push stream failed
 *
 * @return
 * - = 0: Success
 * - < 0: Failure
 * 
 * @note If start success, 'on_push_start' will be triggered
 */
extern __alirts_api_ int alirts_start(alirts_handler_t *handler, const alirts_config_t *config);

/**
 * Send audio and video frame
 *
 * @param[in] handler alirts handler
 * @param[in] frame   audio and video frame, see#alirts_frame_t
 *
 * @return
 * - = 0: Success
 * - < 0: Failure
 */
extern __alirts_api_ int alirts_send_frame(const alirts_handler_t *handler, const alirts_frame_t *frame);

/**
 * Get push statistics info
 *
 * @param[in]  handler alirts handler
 * @param[out] info  push statistics info, see#alirts_push_info_t
 *
 * @return
 * - = 0: Success
 * - < 0: Failure
 */
extern __alirts_api_ int alirts_get_statistics_info(const alirts_handler_t *handler, alirts_push_info_t *info);

/**
 * Stop push stream
 *
 * @param[in] handler alirts handler
 *
 * @note If stop success, 'on_push_stop' will be triggered
 */
extern __alirts_api_ void alirts_stop(alirts_handler_t *handler);

/**
 * Destory the alirts service
 *
 * @param[in] handler alirts handler
 *
 */
extern __alirts_api_ void alirts_destroy(const alirts_handler_t *handler);

#ifdef __cplusplus
}
#endif

#endif // __ALI_RTS_SDK_H
