/******************************************************************************
    Copyright (C) 2021 by aliyun

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
******************************************************************************/

#include "artc-stream.h"
#include <util/threading.h>
#include <util/platform.h>
#include <ali-rts/hermes/src/ali_rts_sdk/include/ali_rts_sdk.h>
#include <jansson.h>
#include <stdio.h>

struct artc_stream {
	obs_output_t *output;
	volatile bool conneted;
	int drop_frames;
	struct dstr url;
	// artc 自定义配置
	struct dstr config_key;
	alirts_config_t config;
	alirts_handler_t* alirts;
};

static const char *artc_stream_getname(void *unused)
{
	UNUSED_PARAMETER(unused);
	return obs_module_text("ArtcStream");
}

static void artc_stream_destroy(void *data)
{
	blog(LOG_INFO, "artc_stream_destroy");
	int64_t start_time = os_gettime_ns();
	struct artc_stream *stream = data;
	if (stream->alirts) {
		alirts_destroy(stream->alirts);
	}
	dstr_free(&stream->url);
	dstr_free(&stream->config_key);
	bfree(stream);
	blog(LOG_INFO, "artc_stream_destroy cost %llu ms", (os_gettime_ns() - start_time) / 1000000);
}

static void set_output_error(struct artc_stream *stream, const char* msg, int code) {
	obs_output_set_last_error(stream->output, msg);
	obs_output_signal_stop(stream->output, code);
	os_atomic_set_bool(&stream->conneted, false);
}

static void engine_stop(struct artc_stream *stream) {
	os_atomic_set_bool(&stream->conneted, false);
	if (stream->alirts) {
		alirts_stop(stream->alirts);
	}
}

// implement alirts event handler callback
// 日志回调，通过该回调可以回调sdk日志，否则sdk默认打印
static void on_log_data(int level,  const char *log) {
	blog(LOG_INFO, "%s", log);
}

// 错误回调
static void on_error(void* user_data, int code, const char *msg) {
	struct artc_stream *stream = user_data;
	const char *str = NULL;
	// blog(LOG_INFO, "engine recv msg: %d\n", type);
	switch (code) {
	case VIDEO_ENCODER_INIT_ERR:
	case VIDEO_ENCODER_ENCODE_ERR: {
		engine_stop(stream);
		str = obs_module_text("VideoEncoderError");
		set_output_error(stream, str, OBS_OUTPUT_ENCODE_ERROR);
		break;
	}
	case AUDIO_ENCODER_INIT_ERR:
	case AUDIO_ENCODER_ENCODE_ERR: {
		engine_stop(stream);
		str = obs_module_text("AudioEncoderError");
		set_output_error(stream, str, OBS_OUTPUT_ENCODE_ERROR);
		break;
	}
	case CONFIG_PARAM_ERR: {
		engine_stop(stream);
		str = obs_module_text("ConfigParamError");
		set_output_error(stream, str, OBS_VIDEO_INVALID_PARAM);
		break;
	}
	default:
		break;
	}
}

// 推流成功，可以发送音视频数据
void on_push_start(void* user_data) {
	struct artc_stream *stream = user_data;
	if (!stream) {
		return;
	}
	os_atomic_set_bool(&stream->conneted, true);
	blog(LOG_INFO, "on_push_start");
}

// 推流结束
void on_push_stop(void* user_data) {
	blog(LOG_INFO, "on_push_stop");
}

// 网络断开
void on_connection_lost(void* user_data) {
	blog(LOG_INFO, "on_connection_lost");
	struct artc_stream *stream = user_data;
	if (!stream)  {
		return;
	}
	os_atomic_set_bool(&stream->conneted, false);
}

// 网络重连状态
void on_reconnect_state(void* user_data, bool success) {
	blog(LOG_INFO, "on_reconnect_state %d", success);
	struct artc_stream *stream = user_data;
	if (!stream)  {
		return;
	}
	if (success) { // 重连成功
		os_atomic_set_bool(&stream->conneted, true);
	} else { // 重连失败
		engine_stop(stream);
		const char *str = obs_module_text("ConnectionLost");
		set_output_error(stream, str, OBS_OUTPUT_CONNECT_FAILED);
	}
}

static void *artc_stream_create(obs_data_t *settings, obs_output_t *output)
{
	blog(LOG_INFO, "artc_stream_create");
	int64_t start_time = os_gettime_ns();
	struct artc_stream *stream = bzalloc(sizeof(struct artc_stream));
	stream->output = output;

	UNUSED_PARAMETER(settings);
    alirts_set_log_level(ALIRTS_LOG_DEBUG);
  // 事件回调设置
	alirts_event_handler_t event_handler;
	event_handler.user_data = stream;
	event_handler.on_log_data = on_log_data;
	event_handler.on_error = on_error;
	event_handler.on_push_start = on_push_start;
	event_handler.on_push_stop  = on_push_stop;
	event_handler.on_connection_lost = on_connection_lost;
	event_handler.on_reconnect_state = on_reconnect_state;
	stream->alirts = alirts_create(&event_handler);
	blog(LOG_INFO, "artc_stream_create cost %llu ms", (os_gettime_ns() - start_time) / 1000000);
	return stream;
}

static void artc_stream_stop(void *data, uint64_t ts)
{
	UNUSED_PARAMETER(ts);
	blog(LOG_INFO, "artc_stream_stop");
	int64_t start_time = os_gettime_ns();
	struct artc_stream *stream = data;
	obs_output_end_data_capture(stream->output);
	engine_stop(stream);
	obs_output_signal_stop(stream->output, OBS_OUTPUT_SUCCESS);
	blog(LOG_INFO, "artc_stream_stop cost %llu ms", (os_gettime_ns() - start_time) / 1000000);
}

// 解析artc自定义配置
// {"use_265":false, "preset":"fast", "profile":"baseline", "enable_auto_reconnect":true, "reconnect_timeout":10}
static void prase_config_key(struct artc_stream *stream, const char* config) {
	// 设置编码器默认配置
	stream->config.video_codec = VIDEO_CODEC_H264;
	stream->config.video_preset = VIDEO_PRESET_VERYFAST;
	stream->config.video_profile = VIDEO_PROFILE_BASELINE;
	// 重连配置
	stream->config.enable_auto_reconnect = true; //建议开启自动重连
	stream->config.reconnect_timeout = 10; // 设置重连超时时间为10s

	if (!config) {
		return;
	}
	blog(LOG_INFO, "config_key %s", config);

	json_t *root = json_loads(config, 0, NULL);
	if (!root) {
        blog(LOG_WARNING, "invaild json, please check it out!");
		return;
	}

	// parse use_265
	json_t *use_h265 = json_object_get(root, "use_265");
	if (use_h265 && json_is_boolean(use_h265)) {
		stream->config.video_codec = json_is_true(use_h265) ? VIDEO_CODEC_H265 : VIDEO_CODEC_H264;
	}

	// parse preset
	json_t *preset = json_object_get(root, "preset");
	if (preset && json_is_string(preset)) {
		const char *preset_name = json_string_value(preset);
		if (strcmp(preset_name, "veryfast") == 0) {
			stream->config.video_preset = VIDEO_PRESET_VERYFAST;
		} else if (strcmp(preset_name, "fast") == 0) {
			stream->config.video_preset = VIDEO_PRESET_FAST;
		} else if (strcmp(preset_name, "medium") == 0) {
			stream->config.video_preset = VIDEO_PRESET_MEDIUM;
		} else if (strcmp(preset_name, "slow") == 0) {
			stream->config.video_preset = VIDEO_PRESET_SLOW;
		}
	}

	// parse profile
	json_t *profile = json_object_get(root, "profile");
	if (profile && json_is_string(profile)) {
		const char *profile_name = json_string_value(profile);
		if (strcmp(profile_name, "baseline") == 0) {
			stream->config.video_profile = VIDEO_PROFILE_BASELINE;
		} else if (strcmp(profile_name, "main") == 0) {
			stream->config.video_profile = VIDEO_PROFILE_MAIN;
		} else if (strcmp(profile_name, "high") == 0) {
			stream->config.video_profile = VIDEO_PROFILE_HIGH;
		}
	}

	// parse enable_auto_reconnect
	json_t *enable_auto_reconnect = json_object_get(root, "enable_auto_reconnect");
	if (enable_auto_reconnect && json_is_boolean(enable_auto_reconnect)) {
		stream->config.enable_auto_reconnect = json_is_true(enable_auto_reconnect);
	}

	// parse reconnect_timeout
	json_t *reconnect_timeout = json_object_get(root, "reconnect_timeout");
	if (reconnect_timeout && json_is_integer(reconnect_timeout)) {
		int timeout = (int)json_integer_value(reconnect_timeout);
		stream->config.reconnect_timeout = timeout > 0 ? timeout : 10;
	}

	blog(LOG_INFO, "codec %d preset %d profile %d enable_auto_reconnect %d reconnect_timeout %d", 
			stream->config.video_codec, stream->config.video_preset, stream->config.video_profile,
			stream->config.enable_auto_reconnect, stream->config.reconnect_timeout);
}

static bool artc_stream_start(void *data)
{
	blog(LOG_INFO, "artc_stream_start");
	int64_t start_time = os_gettime_ns();
	struct artc_stream *stream = data;
	// * Sets the current audio/video media contexts associated with this output,
	// * required for non-encoded outputs.
	obs_output_set_media(stream->output, obs_get_video(), obs_get_audio());

	if (!obs_output_can_begin_data_capture(stream->output, 0)) {
		blog(LOG_ERROR, "obs start capture failed");
		return false;
	}

	// get video info
	struct obs_video_info ovi;
	obs_get_video_info(&ovi);
	enum video_format format = ovi.output_format;
	alirts_config_t *config = &stream->config;

	config->width = ovi.output_width;
	config->height  = ovi.output_height;
	config->max_fps = (float)ovi.fps_num /  (float) ovi.fps_den;

	// Encoder only support yuv i420 formate
	if (format != VIDEO_FORMAT_I420) {
		struct video_scale_info to;
		to.format = VIDEO_FORMAT_I420;
		to.width = config->width;
		to.height = config->height;
		obs_output_set_video_conversion(stream->output, &to);
	}

	struct obs_audio_info oai;
	obs_get_audio_info(&oai);
	config->channel = get_audio_channels(oai.speakers);
	// 音频仅支持44.1k和48k
	config->sample_rate = oai.samples_per_sec;
	// 音频仅支持16bit
	config->sample_format = FORMAT_16BIT;

	blog(LOG_INFO, "audio samplerate %d  speakers %d", config->sample_rate, oai.speakers);
	// 音频仅支持44.1k和48k
	if (!(config->sample_rate == 48000 || config->sample_rate == 44100)) {
		blog(LOG_ERROR, "Only support samplerate 48000 and 44100");
		const char *msg = obs_module_text("unsupport sample rate");
		set_output_error(stream, msg, OBS_OUTPUT_UNSUPPORTED);
		return false;
	}

	// convert audio formate to AUDIO_FORMAT_16BIT
	struct audio_convert_info conversion;
	conversion.format = AUDIO_FORMAT_16BIT;
	conversion.speakers = oai.speakers;
	conversion.samples_per_sec = oai.samples_per_sec;
	obs_output_set_audio_conversion(stream->output, &conversion);


	obs_service_t *service = obs_output_get_service(stream->output);
	if (!service)
		return false;
	dstr_copy(&stream->url, obs_service_get_url(service));
	dstr_copy(&stream->config_key, obs_service_get_key(service));
	dstr_depad(&stream->url);
	dstr_depad(&stream->config_key);
	blog(LOG_INFO, "url %s", stream->url.array);

	// 解析artc输入自定义配置
	prase_config_key(stream, stream->config_key.array);

	obs_encoder_t *venc = obs_output_get_video_encoder(stream->output);
	obs_encoder_t *aenc = obs_output_get_audio_encoder(stream->output, 0);
	obs_data_t *vsettings = obs_encoder_get_settings(venc);
	obs_data_t *asettings = obs_encoder_get_settings(aenc);
    
	int bitrate_kbps = (int)obs_data_get_int(asettings, "bitrate");
	config->audio_bitrate_bps = bitrate_kbps * 1000;
	bitrate_kbps = (int)obs_data_get_int(vsettings, "bitrate");
	config->video_bitrate_bps = bitrate_kbps * 1000;
	int keyint_sec = (int)obs_data_get_int(vsettings, "keyint_sec");
	config->gop_size = (keyint_sec == 0 ? 2 : keyint_sec);

	blog(LOG_INFO, "video width %d height %d fps %d gop %d video %d bps audio %d bps", config->width, config->height,
					config->max_fps, config->gop_size, config->video_bitrate_bps, config->audio_bitrate_bps);

	// 设置推流url
	config->url = stream->url.array;
	// 开始推流
	uint32_t ret = alirts_start(stream->alirts, &stream->config);
	if (ret != 0) {
	blog(LOG_ERROR, "Failed to alirts_start");
		return false;
	}

	obs_output_begin_data_capture(stream->output, 0);
	blog(LOG_INFO, "artc_stream_start cost %llu ms", (os_gettime_ns() - start_time) / 1000000);
	return true;
}

static void receive_video(void *data, struct video_data *frame)
{
	struct artc_stream *stream = data;
	if (!os_atomic_load_bool(&stream->conneted)) {
		return;
	}

	// convert time from ns to ms
	uint64_t pts = frame->timestamp / 1000000;

	alirts_frame_t video_frame;
	video_frame.type = MEDIA_TYPE_VIDEO_I420; // now only this formate
	video_frame.width = stream->config.width;
	video_frame.height = stream->config.height;
	video_frame.timestamp = pts;
	video_frame.data[0] = frame->data[0];
	video_frame.data[1] = frame->data[1];
	video_frame.data[2] = frame->data[2];
	video_frame.linesize[0] = frame->linesize[0];
	video_frame.linesize[1] = frame->linesize[1];
	video_frame.linesize[2] = frame->linesize[2];
	int ret = alirts_send_frame(stream->alirts, &video_frame);
	if (ret != 0) {
		blog(LOG_ERROR, "Artc add video frame error %d", ret);
	}
}

static void receive_audio(void *data, struct audio_data *frame)
{
	struct artc_stream *stream = data;
	if (!os_atomic_load_bool(&stream->conneted)) {
		return;
	}

	uint64_t pts = frame->timestamp / 1000000;

	alirts_frame_t pcm_frame;
	pcm_frame.type = MEDIA_TYPE_AUDIO_PCM;
	pcm_frame.audio_channels = stream->config.channel;
	pcm_frame.audio_nb_samples = frame->frames;
	pcm_frame.audio_sample_rate = stream->config.sample_rate;
	pcm_frame.timestamp = pts;
	pcm_frame.data[0] = frame->data[0];

	int ret = alirts_send_frame(stream->alirts, &pcm_frame);
	if (ret != 0) {
		blog(LOG_ERROR, "Artc add audio frame error %d", ret);
	}
}

uint64_t artc_stream_total_bytes_sent(void* data) {
	struct artc_stream *stream = data;
	alirts_push_info_t info;
	alirts_get_statistics_info(stream->alirts, &info);
	return info.total_send_bytes;
}

int artc_stream_dropped_frames(void *data) {
	struct artc_stream *stream = data;
	return stream->drop_frames;
}

struct obs_output_info artc_output_info = {
	.id = "artc_output",
	.flags =  OBS_OUTPUT_AUDIO | OBS_OUTPUT_VIDEO | OBS_OUTPUT_SERVICE |
			OBS_OUTPUT_CAN_PAUSE,
	.get_name = artc_stream_getname,
	.create = artc_stream_create,
	.destroy = artc_stream_destroy,
	.start = artc_stream_start,
	.stop = artc_stream_stop,
	.raw_video = receive_video,
	.raw_audio = receive_audio,
	.get_total_bytes = artc_stream_total_bytes_sent,
	.get_dropped_frames = artc_stream_dropped_frames,
};
